/*
 * Copyright (c) 2019 AlexIIL
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package alexiil.mc.lib.net.impl;

import net.fabricmc.fabric.api.network.PacketContext;
import net.minecraft.class_156;
import net.minecraft.class_2596;
import net.minecraft.class_2658;
import net.minecraft.class_3244;
import alexiil.mc.lib.net.EnumNetSide;
import alexiil.mc.lib.net.NetByteBuf;

/** A connection on the server side to a specific {@link ServerPlayerEntity}. */
public class ActiveServerConnection extends ActiveMinecraftConnection {
    public final class_3244 netHandler;

    private long serverTick = Long.MIN_VALUE;

    public ActiveServerConnection(class_3244 netHandler) {
        // PacketContext is implemented through the mixin MixinServerPlayNetworkHandler
        super((PacketContext) netHandler);
        this.netHandler = netHandler;
    }

    @Override
    protected class_2596<?> toNormalPacket(NetByteBuf data) {
        return new class_2658(PACKET_ID, data);
    }

    @Override
    protected class_2596<?> toCompactPacket(int receiverId, NetByteBuf data) {
        byte[] bytes = new byte[data.readableBytes()];
        data.readBytes(bytes);
        return new CompactDataPacketToClient(receiverId, bytes);
    }

    @Override
    protected void sendPacket(class_2596<?> packet) {
        netHandler.method_14364(packet);
    }

    @Override
    public EnumNetSide getNetSide() {
        return EnumNetSide.SERVER;
    }

    @Override
    public String toString() {
        return "{ActiveServerConnection for " + netHandler.field_14140 + "}";
    }

    /** @return The value that the client will use to identify the current server tick. */
    public long getServerTick() {
        if (serverTick == Long.MIN_VALUE) {
            serverTick = 0;
        }
        return serverTick;
    }

    @Override
    protected void sendTickPacket() {
        super.sendTickPacket();
        getServerTick();
        if (serverTick != Long.MIN_VALUE) {
            final long sv = serverTick;
            final long now = class_156.method_658();
            NET_ID_SERVER_TICK.send(this, (buffer, ctx) -> {
                buffer.writeLong(sv);
                buffer.writeLong(now);
            });
            serverTick++;
            if (serverTick < 0) {
                // Overflow, this probably won't end well as we use min_value as the sentinal value.
                serverTick++;
            }
        }
    }
}
