/*
 * Copyright (c) 2019 AlexIIL
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package alexiil.mc.lib.net.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import javax.annotation.Nullable;

import net.fabricmc.fabric.api.event.client.ClientTickCallback;
import net.fabricmc.fabric.api.event.server.ServerStopCallback;
import net.fabricmc.fabric.api.event.server.ServerTickCallback;
import net.fabricmc.fabric.api.network.ClientSidePacketRegistry;
import net.fabricmc.fabric.api.network.PacketContext;
import net.fabricmc.fabric.api.network.ServerSidePacketRegistry;
import net.fabricmc.fabric.api.server.PlayerStream;
import net.minecraft.class_1657;
import net.minecraft.class_1937;
import net.minecraft.class_2338;
import net.minecraft.class_2539;
import net.minecraft.class_2585;
import net.minecraft.class_2586;
import net.minecraft.class_2598;
import net.minecraft.class_310;
import net.minecraft.class_3222;
import net.minecraft.class_3244;
import net.minecraft.class_634;
import net.minecraft.class_746;
import alexiil.mc.lib.net.InvalidInputDataException;
import alexiil.mc.lib.net.LibNetworkStack;
import alexiil.mc.lib.net.NetByteBuf;
import alexiil.mc.lib.net.mixin.api.INetworkStateMixin;

public class CoreMinecraftNetUtil {

    private static final boolean DEBUG = LibNetworkStack.DEBUG;

    static int serverExpectedId;
    static int clientExpectedId;

    private static ActiveClientConnection currentClientConnection;
    private static final Map<PacketContext, ActiveServerConnection> serverConnections = new HashMap<>();

    public static List<ActiveMinecraftConnection> getNearbyActiveConnections(class_2586 be, int distance) {
        List<ActiveMinecraftConnection> list = new ArrayList<>();
        class_1937 w = be.method_10997();
        if (w == null) {
            return list;
        }
        if (w.field_9236) {
            if (currentClientConnection == null) {
                return list;
            }
            double distanceSq = distance * distance;
            double x = be.method_11016().method_10263() + 0.5;
            double y = be.method_11016().method_10264() + 0.5;
            double z = be.method_11016().method_10260() + 0.5;
            if (currentClientConnection.getMinecraftContext().getPlayer().method_5649(x, y, z) < distanceSq) {
                list.add(currentClientConnection);
            }
        } else {
            Set<class_1657> players = PlayerStream.around(w, be.method_11016(), distance).collect(Collectors.toSet());
            for (class_1657 player : players) {
                list.add(getServerConnection((PacketContext) ((class_3222) player).field_13987));
            }
        }
        return list;
    }

    public static List<ActiveMinecraftConnection> getPlayersWatching(class_1937 world, class_2338 pos) {
        List<class_1657> players = PlayerStream.watching(world, pos).collect(Collectors.toList());
        List<ActiveMinecraftConnection> list = new ArrayList<>();
        for (class_1657 player : players) {
            list.add(getConnection(player));
        }
        return list;
    }

    public static ActiveMinecraftConnection getConnection(class_1657 player) {
        if (player == null) {
            throw new NullPointerException("player");
        }
        if (player instanceof class_3222) {
            return getServerConnection((PacketContext) ((class_3222) player).field_13987);
        } else if (player.field_6002.field_9236) {
            class_746 clientPlayer = class_310.method_1551().field_1724;
            if (clientPlayer == null) {
                throw new IllegalStateException(
                    "Cannot create a connection for " + player + " because they cannot be the client player (null)!"
                );
            }
            if (currentClientConnection == null && player == clientPlayer) {
                return getOrCreateClientConnection((PacketContext) clientPlayer.field_3944);
            } else if (currentClientConnection != null && currentClientConnection.ctx.getPlayer() == player) {
                return currentClientConnection;
            } else {
                throw new IllegalArgumentException("Unknown PlayerEntity " + player.getClass());
            }
        } else {
            throw new IllegalArgumentException("Unknown PlayerEntity " + player.getClass());
        }
    }

    public static void load() {
        ServerSidePacketRegistry.INSTANCE.register(ActiveMinecraftConnection.PACKET_ID, (ctx, buffer) -> {
            onServerReceivePacket(ctx, NetByteBuf.asNetByteBuf(buffer));
        });

        ServerTickCallback.EVENT.register(server -> onServerTick());
        ServerStopCallback.EVENT.register(server -> onServerStop());

        INetworkStateMixin play = (INetworkStateMixin) (Object) class_2539.field_20591;
        clientExpectedId = play.libnetworkstack_registerPacket(
            class_2598.field_11942, CompactDataPacketToClient.class, CompactDataPacketToClient::new
        );
        serverExpectedId = play.libnetworkstack_registerPacket(
            class_2598.field_11941, CompactDataPacketToServer.class, CompactDataPacketToServer::new
        );
    }

    public static void loadClient() {
        ClientSidePacketRegistry.INSTANCE.register(ActiveMinecraftConnection.PACKET_ID, (ctx, buffer) -> {
            onClientReceivePacket(ctx, NetByteBuf.asNetByteBuf(buffer));
        });

        ClientTickCallback.EVENT.register(client -> onClientTick());
    }

    static void onClientReceivePacket(PacketContext ctx, NetByteBuf buffer) {
        ActiveClientConnection connection = getOrCreateClientConnection(ctx);
        NetByteBuf b = buffer.copy();
        ctx.getTaskQueue().execute(() -> {
            try {
                connection.onReceiveRawData(b);
                b.release();
            } catch (InvalidInputDataException e) {
                e.printStackTrace();
                ((class_634) ctx).method_2872()
                    .method_10747(new class_2585("LibNetworkStack: read error (see logs for details)"));
            }
        });
    }

    /** @return The current client connection. Throws an exception if this is not available right now (likely because
     *         the client hasn't joined a world). */
    public static ActiveClientConnection getClientConnection() {
        class_746 player = class_310.method_1551().field_1724;
        if (player == null) {
            throw new IllegalStateException("The client is not currently connected to any server!");
        }
        return getOrCreateClientConnection((PacketContext) player.field_3944);
    }

    /** @return The current {@link ActiveClientConnection}, or null if one has not been created yet. */
    @Nullable
    public static ActiveClientConnection getCurrentClientConnection() {
        return currentClientConnection;
    }

    private static ActiveClientConnection getOrCreateClientConnection(PacketContext ctx) {
        ActiveClientConnection connection = currentClientConnection;
        if (connection == null || connection.getMinecraftContext() != ctx) {
            connection = new ActiveClientConnection((class_634) ctx);
            connection.postConstruct();
            currentClientConnection = connection;
        }
        return connection;
    }

    static void onServerReceivePacket(PacketContext ctx, NetByteBuf buffer) {
        ActiveServerConnection connection = getServerConnection(ctx);
        if (connection == null) {
            return;
        }

        NetByteBuf b = buffer.copy();
        ctx.getTaskQueue().execute(() -> {
            try {
                connection.onReceiveRawData(b);
                b.release();
            } catch (InvalidInputDataException e) {
                e.printStackTrace();
                ((class_3244) ctx).method_14367(
                    new class_2585("LibNetworkStack: read error (see server logs for more details)\n" + e)
                );
            }
        });
    }

    private static ActiveServerConnection getServerConnection(PacketContext ctx) {
        return serverConnections.computeIfAbsent(ctx, c -> {
            class_3244 networkHandler = (class_3244) c;
            if (!networkHandler.field_14127.method_10758()) {
                if (DEBUG) {
                    LibNetworkStack.LOGGER.info(
                        "Disallowed server connection for " + ctx.getPlayer() + " because it's channel is not open!"
                    );
                }
                return null;
            } else {
                if (DEBUG) {
                    LibNetworkStack.LOGGER.info("Creating new server connection for " + ctx.getPlayer());
                }
            }
            ActiveServerConnection connection = new ActiveServerConnection(networkHandler);
            connection.postConstruct();
            return connection;
        });
    }

    private static void onClientTick() {
        if (currentClientConnection != null) {
            if (!currentClientConnection.netHandler.method_2872().method_10758()) {
                if (DEBUG) {
                    LibNetworkStack.LOGGER.info(
                        "Removed current client connection " + currentClientConnection
                            + " as it's channel is no longer open."
                    );
                }
                currentClientConnection = null;
                return;
            }
            currentClientConnection.tick();
        }
    }

    private static void onServerTick() {
        Iterator<ActiveServerConnection> iterator = serverConnections.values().iterator();
        while (iterator.hasNext()) {
            ActiveServerConnection connection = iterator.next();
            if (!connection.netHandler.field_14127.method_10758()) {
                iterator.remove();
                if (DEBUG) {
                    LibNetworkStack.LOGGER
                        .info("Removed server connection " + connection + " as it's channel is no longer open.");
                }
                continue;
            }
            connection.tick();
        }
    }

    private static void onServerStop() {
        int count = serverConnections.size();
        serverConnections.clear();
        if (DEBUG) {
            LibNetworkStack.LOGGER.info("Removed " + count + " server connections as the server has stopped.");
        }
    }
}
