/*
 * Copyright (c) 2019 AlexIIL
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at https://mozilla.org/MPL/2.0/.
 */
package alexiil.mc.lib.net.impl;

import java.util.Objects;
import net.minecraft.class_1297;
import net.minecraft.class_1657;
import net.minecraft.class_1703;
import net.minecraft.class_1792;
import net.minecraft.class_1799;
import net.minecraft.class_2487;
import net.minecraft.class_2586;
import alexiil.mc.lib.net.ActiveConnection;
import alexiil.mc.lib.net.IMsgReadCtx;
import alexiil.mc.lib.net.IMsgWriteCtx;
import alexiil.mc.lib.net.InvalidInputDataException;
import alexiil.mc.lib.net.NetByteBuf;
import alexiil.mc.lib.net.NetObjectCache;
import alexiil.mc.lib.net.NetObjectCache.IEntrySerialiser;
import alexiil.mc.lib.net.ParentNetId;
import alexiil.mc.lib.net.ParentNetIdSingle;

import it.unimi.dsi.fastutil.Hash;

/** Holder for everything related to a normal minecraft connection. */
public final class McNetworkStack {
    private McNetworkStack() {}

    /** The root parent - everything must use as it's parent (somewhere up the chain) or it can never be written out to
     * a minecraft connection. */
    public static final ParentNetId ROOT = new ParentNetId(null, "");

    public static final ParentNetIdSingle<class_2586> BLOCK_ENTITY;
    public static final ParentNetIdSingle<class_1297> ENTITY;

    public static final ParentNetIdSingle<class_1703> SCREEN_HANDLER;

    /** @deprecated Use {@link #SCREEN_HANDLER} instead, due to the yarn name change. */
    @Deprecated
    public static final ParentNetIdSingle<class_1703> CONTAINER;

    public static final NetObjectCache<class_1799> CACHE_ITEMS_WITHOUT_AMOUNT;

    static {
        BLOCK_ENTITY = new ParentNetIdSingle<class_2586>(ROOT, class_2586.class, "block_entity", -1) {
            @Override
            public class_2586 readContext(NetByteBuf buffer, IMsgReadCtx ctx) throws InvalidInputDataException {
                ActiveMinecraftConnection mcConn = (ActiveMinecraftConnection) ctx.getConnection();
                class_1657 player = mcConn.ctx.getPlayer();
                return player.field_6002.method_8321(buffer.method_10811());
            }

            @Override
            public void writeContext(NetByteBuf buffer, IMsgWriteCtx ctx, class_2586 value) {
                buffer.writeBlockPos(value.method_11016());
            }
        };
        ENTITY = new ParentNetIdSingle<class_1297>(ROOT, class_1297.class, "entity", Integer.BYTES) {
            @Override
            public class_1297 readContext(NetByteBuf buffer, IMsgReadCtx ctx) throws InvalidInputDataException {
                ActiveMinecraftConnection mcConn = (ActiveMinecraftConnection) ctx.getConnection();
                class_1657 player = mcConn.ctx.getPlayer();
                return player.field_6002.method_8469(buffer.readInt());
            }

            @Override
            public void writeContext(NetByteBuf buffer, IMsgWriteCtx ctx, class_1297 value) {
                buffer.writeInt(value.method_5628());
            }
        };
        // We're not changing the internal name, because that would needlessly break compat
        SCREEN_HANDLER = new ParentNetIdSingle<class_1703>(ROOT, class_1703.class, "container", Integer.BYTES) {
            @Override
            public class_1703 readContext(NetByteBuf buffer, IMsgReadCtx ctx) throws InvalidInputDataException {
                ActiveMinecraftConnection mcConn = (ActiveMinecraftConnection) ctx.getConnection();
                class_1657 player = mcConn.ctx.getPlayer();
                int syncId = buffer.readInt();
                class_1703 currentContainer = player.field_7512;
                if (currentContainer == null || currentContainer.field_7763 != syncId) {
                    return null;
                }
                return currentContainer;
            }

            @Override
            public void writeContext(NetByteBuf buffer, IMsgWriteCtx ctx, class_1703 value) {
                buffer.writeInt(value.field_7763);
            }
        };
        CONTAINER = SCREEN_HANDLER;

        CACHE_ITEMS_WITHOUT_AMOUNT
            = new NetObjectCache<>(ROOT.child("cache_item_stack"), new Hash.Strategy<class_1799>()
            {

                @Override
                public int hashCode(class_1799 o) {
                    return Objects.hash(new Object[] { o.method_7909(), o.method_7969() });
                }

                @Override
                public boolean equals(class_1799 a, class_1799 b) {
                    if (a == null || b == null) {
                        return a == b;
                    }
                    if (a.method_7909() != b.method_7909()) {
                        return false;
                    }
                    class_2487 tagA = a.method_7969();
                    class_2487 tagB = b.method_7969();
                    if (tagA == null || tagB == null) {
                        return tagA == tagB;
                    }
                    return tagA.equals(tagB);
                }

            }, new IEntrySerialiser<class_1799>() {

                @Override
                public void write(class_1799 obj, ActiveConnection connection, NetByteBuf buffer) {
                    if (obj.method_7960()) {
                        buffer.writeBoolean(false);
                    } else {
                        buffer.writeBoolean(true);
                        class_1792 item_1 = obj.method_7909();
                        buffer.writeVarInt(class_1792.method_7880(item_1));
                        class_2487 tag = null;
                        if (item_1.method_7846() || item_1.method_7887()) {
                            tag = obj.method_7969();
                        }
                        buffer.method_10794(tag);
                    }
                }

                @Override
                public class_1799 read(ActiveConnection connection, NetByteBuf buffer) {
                    if (!buffer.readBoolean()) {
                        return class_1799.field_8037;
                    } else {
                        int id = buffer.method_10816();
                        class_1799 stack = new class_1799(class_1792.method_7875(id));
                        stack.method_7980(buffer.method_10798());
                        return stack;
                    }
                }
            });
    }
}
